#define _GNU_SOURCE /* strchrnul関数使用のために必要 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#define MAIN_C
#include "exitfail.h"

/* 表示する文字数 */
#define DISP_WIDTH  60 /* 幅 */
#define DISP_HEIGHT 17 /* 高さ */

/* CSVデータ各要素のサイズ */
#define ZIPCODE_LEN 8
#define STREET_LEN  512
#define CITY_LEN    64
#define PREF_LEN    16
#define FLAG_NUM    6
/* CSVデータ1行の要素数 */
#define COLUMN_NUM  (5 + FLAG_NUM)

/* CSVデータ構造体 */
typedef struct {
	long code;                 /* 全国地方公共団体コード */
	char zipcode[ZIPCODE_LEN]; /* 郵便番号 */
	char street[STREET_LEN];   /* 町域名 */
	char city[CITY_LEN];       /* 市区町村名 */
	char pref[PREF_LEN];       /* 都道府県名 */
	long flag[FLAG_NUM];       /* フラグ配列 */
} csvline_t;

#define BASENAME(p) ((strrchr((p), '/') ? : ((p) - 1)) + 1)

/* conf設定ファイル内で使用するキーワードの定義 */
#define GROUP_DISPLAY "dispaly"
#define KEY_WIDTH     "Width"
#define KEY_HEIGHT    "Height"
#define GROUP_CONTROL "control"
#define KEY_PAUSE     "Pause"
#define KEY_COUNT     "Count"
#define GROUP_DATA    "data"
#define KEY_CODE      "Code"
#define KEY_ZIPCODE   "Zipcode"
#define KEY_STREET    "Street"
#define KEY_CITY      "City"
#define KEY_PREF      "Pref"
#define KEY_FLAG      "Flag"

/* conf設定保持用構造体 */
typedef struct {
	struct { /* 表示設定グループ */
		gint width;  /* 表示幅 */
		gint height; /* 表示高さ */
	} display;
	struct { /* 制御設定グループ */
		gboolean pause; /* 一時停止の有無 */
		gboolean count; /* 総数表示の有無 */
	} control;
	struct { /* データ条件設定グループ */
		gint code;      /* 全国地方公共団体コード */
		gchar *zipcode; /* 郵便番号(部分一致) */
		gchar *street;  /* 町域名(部分一致) */
		gchar *city;    /* 市区町村名(部分一致) */
		gchar *pref;    /* 都道府県名(部分一致) */
		gint *flag;     /* フラグ配列 */
		gsize flag_len; /* フラグ配列要素数 */
	} data;
} config_t;
/* conf設定保持用領域ポインタ */
static config_t *pconf;

/**
 * confファイル読み込み関数
 * @param conffilename confファイル名
 */
static void readconf(char conffilename[])
{
	GKeyFile *keyfile;
	GError *error = NULL;
	gint flag[FLAG_NUM];
	gsize len;
	gchar *pdata;
	FILE *pconffile;
	int i;

	/* conf設定保持用領域確保 */
	pconf = g_slice_new(config_t);

	/* キーファイル確保 */
	keyfile = g_key_file_new();

	/* confからキーファイル読み込み、失敗した場合は新規作成 */
	if (!g_key_file_load_from_file(keyfile, conffilename,
				       G_KEY_FILE_KEEP_COMMENTS |
				       G_KEY_FILE_KEEP_TRANSLATIONS,
				       &error)) {
		/* デフォルト設定 */
		pconf->display.width = DISP_WIDTH;
		pconf->display.height = DISP_HEIGHT;
		pconf->control.pause = TRUE;
		pconf->control.count = TRUE;
		pconf->data.code = -1;
		pconf->data.zipcode = "";
		pconf->data.street = "";
		pconf->data.city = "";
		pconf->data.pref = "";
		for (i = 0; i < FLAG_NUM; i++)
			flag[i] = -1;
		pconf->data.flag = flag;
		pconf->data.flag_len = FLAG_NUM;
		/* キーファイル書き込み */
		g_key_file_set_integer(keyfile, GROUP_DISPLAY, KEY_WIDTH,
				       pconf->display.width);
		g_key_file_set_integer(keyfile, GROUP_DISPLAY, KEY_HEIGHT,
				       pconf->display.height);
		g_key_file_set_boolean(keyfile, GROUP_CONTROL, KEY_PAUSE,
				       pconf->control.pause);
		g_key_file_set_boolean(keyfile, GROUP_CONTROL, KEY_COUNT,
				       pconf->control.count);
		g_key_file_set_integer(keyfile, GROUP_DATA, KEY_CODE,
				       pconf->data.code);
		g_key_file_set_string(keyfile, GROUP_DATA, KEY_ZIPCODE,
				      pconf->data.zipcode);
		g_key_file_set_string(keyfile, GROUP_DATA, KEY_STREET,
				      pconf->data.zipcode);
		g_key_file_set_string(keyfile, GROUP_DATA, KEY_CITY,
				      pconf->data.city);
		g_key_file_set_string(keyfile, GROUP_DATA, KEY_PREF,
				      pconf->data.pref);
		g_key_file_set_integer_list(keyfile, GROUP_DATA, KEY_FLAG,
					    pconf->data.flag,
					    pconf->data.flag_len);
		/* キーファイルからテキストデータ取得 */
		pdata = g_key_file_to_data(keyfile, &len, &error);
		if (!pdata)
			g_error(error->message);

		/* 新規ファイルを作成してテキストデータ書き込み */
		pconffile = fopen(conffilename, "w");
		if (!pconffile)
			exitfail_errno("fopen");
		if (fwrite(pdata, len, 1, pconffile) < 1)
			exitfail_errno("fwrite");
		fclose(pconffile);
	}
	/* conf読み込みに成功した場合、設定を保持 */
	else {
		pconf->display.width =
			g_key_file_get_integer(keyfile, GROUP_DISPLAY,
					       KEY_WIDTH, NULL);
		pconf->display.height =
			g_key_file_get_integer(keyfile, GROUP_DISPLAY,
					       KEY_HEIGHT, NULL);
		pconf->control.pause =
			g_key_file_get_boolean(keyfile, GROUP_CONTROL,
					       KEY_PAUSE, NULL);
		pconf->control.count =
			g_key_file_get_boolean(keyfile, GROUP_CONTROL,
					       KEY_COUNT, NULL);
		pconf->data.code =
			g_key_file_get_integer(keyfile, GROUP_DATA,
					       KEY_CODE, NULL);
		pconf->data.zipcode =
			g_key_file_get_string(keyfile, GROUP_DATA,
					      KEY_ZIPCODE, NULL);
		pconf->data.street =
			g_key_file_get_string(keyfile, GROUP_DATA,
					      KEY_STREET, NULL);
		pconf->data.city =
			g_key_file_get_string(keyfile, GROUP_DATA,
					      KEY_CITY, NULL);
		pconf->data.pref =
			g_key_file_get_string(keyfile, GROUP_DATA,
					      KEY_PREF, NULL);
		pconf->data.flag =
			g_key_file_get_integer_list(keyfile, GROUP_DATA,
						    KEY_FLAG,
						    &pconf->data.flag_len,
						    NULL);
	}

	/* キーファイル開放 */
	g_key_file_free(keyfile);
}

/**
 * 行表示関数
 * @param pcsvline 表示するCSVデータ構造体へのポインタ(NULLの場合は表示済データの総数を表示)
 */
static void printline(csvline_t *pcsvline)
{
	static int count = 0;	/* 表示したデータ総数 */
	static int line = 0;	/* 一画面中に表示した行数 */
	int disp_width = DISP_WIDTH, disp_height = DISP_HEIGHT, newline;
	char buf[1024];
	unsigned int i;

	/* CSVデータの各要素を表示 */
	if (pcsvline) {
		/* 郵便番号が入っていない場合は表示しない */
		if (!pcsvline->zipcode[0])
			return;

		/* 全国地方公共団体コード条件設定があり、一致しなかったら表示しない */
		if (pconf->data.code >= 0 && pcsvline->code != pconf->data.code)
			return;
		/* 郵便番号条件設定があり、部分一致しなかったら表示しない */
		if (pconf->data.zipcode[0] != '\0' &&
		    !strcasestr(pcsvline->zipcode, pconf->data.zipcode))
			return;
		/* 町域名条件設定があり、部分一致しなかったら表示しない */
		if (pconf->data.street[0] != '\0' &&
		    !strcasestr(pcsvline->street, pconf->data.street))
			return;
		/* 市区町村名条件設定があり、部分一致しなかったら表示しない */
		if (pconf->data.city[0] != '\0' &&
		    !strcasestr(pcsvline->city, pconf->data.city))
			return;
		/* 都道府県条件設定があり、部分一致しなかったら表示しない */
		if (pconf->data.pref[0] != '\0' &&
		    !strcasestr(pcsvline->pref, pconf->data.pref))
			return;
		for (i = 0; i < FLAG_NUM && i < pconf->data.flag_len; i++) {
			/* フラグ条件設定があり、一致しなかったら表示しない */
			if (pconf->data.flag[i] >= 0 &&
			    pcsvline->flag[i] != pconf->data.flag[i])
				return;
		}

		/* 各要素を表示フォーマットに展開 */
		snprintf(buf, sizeof(buf),
			 "%6d %05ld %s %s %s %s %ld %ld %ld %ld %ld %ld",
			 count++, pcsvline->code, pcsvline->zipcode,
			 pcsvline->street, pcsvline->city, pcsvline->pref,
			 pcsvline->flag[0], pcsvline->flag[1],
			 pcsvline->flag[2], pcsvline->flag[3],
			 pcsvline->flag[4], pcsvline->flag[5]);
	}
	/* CSVデータの総数を表示 */
	else {
		/* 総数表示無効なら表示しない */
		if (!pconf->control.count)
			return;
		/* データ総数を表示フォーマットに展開 */
		sprintf(buf, "Count: %6d", count);
	}

	/* 今回追加される表示行数を計算 */
	disp_width = pconf->display.width;
	disp_height = pconf->display.height;
	newline = (strlen(buf) + (disp_width - 1)) / disp_width;
	/* 1画面を超える場合、入力があるまで一時停止 */
	if (line + newline >= disp_height) {
		/* 一時停止有効なら */
		if (pconf->control.pause)
			getchar();
		/* 表示行数を初期化 */
		line = 0;
	}
	/* 実際に表示する */
	printf("%s\n", buf);
	/* 表示行数を更新 */
	line += newline;
}

/**
 * main関数
 * @param argc 引数なしの場合はusage表示のみ
 * @param argv 第1引数として読み込みCSVファイル名を指定
 * @return exit値
 */
int main(int argc, char *argv[])
{
	FILE *pcsvfile; /* CSVファイルポインタ */
	csvline_t csvline; /* CSVデータ */
	char buf[256], *pbuf, *pcol[COLUMN_NUM];
	int i;

	exitfail_init();

	/* 引数が指定されなかった場合、usage表示して終了 */
	if (argc < 2) {
		printf("Usage: %s <csvfile>\n", BASENAME(argv[0]));
		return EXIT_SUCCESS;
	}

	/* confファイル名を作成 */
	snprintf(buf, sizeof(buf), "%s.conf", argv[0]);
	/* confファイルを読み込み */
	readconf(buf);

	/* CSVファイルオープン */
	pcsvfile = fopen(argv[1], "r");
	if (!pcsvfile)
		exitfail_errno("fopen");

	/* CSVデータを初期化 */
	memset(&csvline, 0, sizeof(csvline));
	/* CSVファイルから1行読み込む */
	while (fgets(buf, sizeof(buf), pcsvfile)) {
		/* 各要素へのポインタ配列を初期化 */
		memset(pcol, 0, sizeof(pcol));
		/* 次のカンマ(または行末)を見つけてトークン化 */
		pbuf = strtok(buf, ",\n");
		for (i = 0; i < COLUMN_NUM && pbuf; i++) {
			/* 前後のダブルクォートは除去する */
			if (*pbuf == '"')
				*strchrnul(++pbuf, '"') = '\0';
			/* 要素へのポインタを保持 */
			pcol[i] = pbuf;
			/* 次のカンマ(または行末)を見つけてトークン化 */
			pbuf = strtok(NULL, ",\n");
		}
		/* 要素数が不足している場合、次行にスキップ */
		if (i < COLUMN_NUM)
			continue;

		/* 新しいデータの場合 */
		if (strcmp(pcol[1], csvline.zipcode) ||
		    strtol(pcol[8], NULL, 10)) {
			/* 保持済みのデータを表示 */
			printline(&csvline);

			/* CSVデータ各要素を保持 */
			memset(&csvline, 0, sizeof(csvline));
			csvline.code = strtol(pcol[0], NULL, 10);
			strncpy(csvline.zipcode, pcol[1],
				sizeof(csvline.zipcode) - 1);
			strncpy(csvline.street, pcol[2],
				sizeof(csvline.street) - 1);
			strncpy(csvline.city, pcol[3],
				sizeof(csvline.city) - 1);
			strncpy(csvline.pref, pcol[4],
				sizeof(csvline.pref) - 1);
			for (i = 0; i < FLAG_NUM; i++)
				csvline.flag[i] = strtol(pcol[5 + i], NULL, 10);
		}
		/* 既存データへの追加の場合 */
		else
			/* 町域名の続きを追加 */
			strncat(csvline.street, pcol[2],
				(sizeof(csvline.street) -
				 strlen(csvline.street)) - 1);
	}
	/* 保持済みのデータを表示 */
	printline(&csvline);

	/* CSVファイルクローズ */
	fclose(pcsvfile);

	/* データ総数を表示 */
	printline(NULL);

	return EXIT_SUCCESS;
}
